@@ -34,9 +34,7 @@
  *  Robert Bragg   <robert@linux.intel.com>
  */
 
-#ifdef HAVE_CONFIG_H
 #include "cogl-config.h"
-#endif
 
 #include "cogl-util.h"
 #include "cogl-bitmap.h"
@@ -48,21 +46,16 @@
 #include "cogl-texture-driver.h"
 #include "cogl-texture-2d-sliced-private.h"
 #include "cogl-texture-2d-private.h"
-#include "cogl-texture-2d-gl.h"
-#include "cogl-texture-3d-private.h"
-#include "cogl-texture-rectangle-private.h"
 #include "cogl-sub-texture-private.h"
 #include "cogl-atlas-texture-private.h"
 #include "cogl-pipeline.h"
 #include "cogl-context-private.h"
 #include "cogl-object-private.h"
 #include "cogl-object-private.h"
-#include "cogl-primitives.h"
 #include "cogl-framebuffer-private.h"
 #include "cogl1-context.h"
 #include "cogl-sub-texture.h"
 #include "cogl-primitive-texture.h"
-#include "cogl-error-private.h"
 #include "cogl-gtype-private.h"
 
 #include <string.h>
@@ -96,7 +89,7 @@ _cogl_texture_register_texture_type (const CoglObjectClass *klass)
   _cogl_texture_types = g_slist_prepend (_cogl_texture_types, (void *) klass);
 }
 
-CoglBool
+gboolean
 cogl_is_texture (void *object)
 {
   CoglObject *obj = (CoglObject *)object;
@@ -122,7 +115,8 @@ _cogl_texture_init (CoglTexture *texture,
                     const CoglTextureVtable *vtable)
 {
   texture->context = context;
-  texture->max_level = 0;
+  texture->max_level_set = 0;
+  texture->max_level_requested = 1000; /* OpenGL default GL_TEXTURE_MAX_LEVEL */
   texture->width = width;
   texture->height = height;
   texture->allocated = FALSE;
@@ -159,7 +153,6 @@ _cogl_texture_free_loader (CoglTexture *texture)
         {
         case COGL_TEXTURE_SOURCE_TYPE_SIZED:
         case COGL_TEXTURE_SOURCE_TYPE_EGL_IMAGE:
-        case COGL_TEXTURE_SOURCE_TYPE_GL_FOREIGN:
         case COGL_TEXTURE_SOURCE_TYPE_EGL_IMAGE_EXTERNAL:
           break;
         case COGL_TEXTURE_SOURCE_TYPE_BITMAP:
@@ -182,10 +175,10 @@ _cogl_texture_free (CoglTexture *texture)
 {
   _cogl_texture_free_loader (texture);
 
-  free (texture);
+  g_free (texture);
 }
 
-CoglBool
+gboolean
 _cogl_texture_needs_premult_conversion (CoglPixelFormat src_format,
                                         CoglPixelFormat dst_format)
 {
@@ -196,13 +189,13 @@ _cogl_texture_needs_premult_conversion (CoglPixelFormat src_format,
           (dst_format & COGL_PREMULT_BIT));
 }
 
-CoglBool
-_cogl_texture_is_foreign (CoglTexture *texture)
+gboolean
+cogl_texture_is_get_data_supported (CoglTexture *texture)
 {
-  if (texture->vtable->is_foreign)
-    return texture->vtable->is_foreign (texture);
+  if (texture->vtable->is_get_data_supported)
+    return texture->vtable->is_get_data_supported (texture);
   else
-    return FALSE;
+    return TRUE;
 }
 
 unsigned int
@@ -237,14 +230,16 @@ _cogl_texture_get_n_levels (CoglTexture *texture)
   int width = cogl_texture_get_width (texture);
   int height = cogl_texture_get_height (texture);
   int max_dimension = MAX (width, height);
+  int n_levels = _cogl_util_fls (max_dimension);
 
-  if (cogl_is_texture_3d (texture))
-    {
-      CoglTexture3D *tex_3d = COGL_TEXTURE_3D (texture);
-      max_dimension = MAX (max_dimension, tex_3d->depth);
-    }
+  return MIN (n_levels, texture->max_level_requested + 1);
+}
 
-  return _cogl_util_fls (max_dimension);
+void
+cogl_texture_set_max_level (CoglTexture *texture,
+                            int          max_level)
+{
+  texture->max_level_requested = max_level;
 }
 
 void
@@ -256,17 +251,9 @@ _cogl_texture_get_level_size (CoglTexture *texture,
 {
   int current_width = cogl_texture_get_width (texture);
   int current_height = cogl_texture_get_height (texture);
-  int current_depth;
+  int current_depth = 0;
   int i;
 
-  if (cogl_is_texture_3d (texture))
-    {
-      CoglTexture3D *tex_3d = COGL_TEXTURE_3D (texture);
-      current_depth = tex_3d->depth;
-    }
-  else
-    current_depth = 0;
-
   /* NB: The OpenGL spec (like D3D) uses a floor() convention to
    * round down the size of a mipmap level when dividing the size
    * of the previous level results in a fraction...
@@ -286,7 +273,7 @@ _cogl_texture_get_level_size (CoglTexture *texture,
     *depth = current_depth;
 }
 
-CoglBool
+gboolean
 cogl_texture_is_sliced (CoglTexture *texture)
 {
   if (!texture->allocated)
@@ -298,7 +285,7 @@ cogl_texture_is_sliced (CoglTexture *texture)
  * will be needed to iterate over multiple sub textures for regions whos
  * texture coordinates extend out of the range [0,1]
  */
-CoglBool
+gboolean
 _cogl_texture_can_hardware_repeat (CoglTexture *texture)
 {
   if (!texture->allocated)
@@ -324,7 +311,7 @@ _cogl_texture_transform_quad_coords_to_gl (CoglTexture *texture,
   return texture->vtable->transform_quad_coords_to_gl (texture, coords);
 }
 
-CoglBool
+gboolean
 cogl_texture_get_gl_texture (CoglTexture *texture,
 			     GLuint *out_gl_handle,
 			     GLenum *out_gl_target)
@@ -336,12 +323,6 @@ cogl_texture_get_gl_texture (CoglTexture *texture,
                                           out_gl_handle, out_gl_target);
 }
 
-CoglTextureType
-_cogl_texture_get_type (CoglTexture *texture)
-{
-  return texture->vtable->get_type (texture);
-}
-
 void
 _cogl_texture_pre_paint (CoglTexture *texture, CoglTexturePrePaintFlags flags)
 {
@@ -368,7 +349,7 @@ _cogl_texture_ensure_non_quad_rendering (CoglTexture *texture)
   texture->vtable->ensure_non_quad_rendering (texture);
 }
 
-CoglBool
+gboolean
 _cogl_texture_set_region_from_bitmap (CoglTexture *texture,
                                       int src_x,
                                       int src_y,
@@ -378,14 +359,12 @@ _cogl_texture_set_region_from_bitmap (CoglTexture *texture,
                                       int dst_x,
                                       int dst_y,
                                       int level,
-                                      CoglError **error)
+                                      GError **error)
 {
-  _COGL_RETURN_VAL_IF_FAIL ((cogl_bitmap_get_width (bmp) - src_x)
-                            >= width, FALSE);
-  _COGL_RETURN_VAL_IF_FAIL ((cogl_bitmap_get_height (bmp) - src_y)
-                            >= height, FALSE);
-  _COGL_RETURN_VAL_IF_FAIL (width > 0, FALSE);
-  _COGL_RETURN_VAL_IF_FAIL (height > 0, FALSE);
+  g_return_val_if_fail (cogl_bitmap_get_width (bmp) - src_x >= width, FALSE);
+  g_return_val_if_fail (cogl_bitmap_get_height (bmp) - src_y >= height, FALSE);
+  g_return_val_if_fail (width > 0, FALSE);
+  g_return_val_if_fail (height > 0, FALSE);
 
   /* Assert that the storage for this texture has been allocated */
   if (!cogl_texture_allocate (texture, error))
@@ -407,7 +386,7 @@ _cogl_texture_set_region_from_bitmap (CoglTexture *texture,
                                       error);
 }
 
-CoglBool
+gboolean
 cogl_texture_set_region_from_bitmap (CoglTexture *texture,
                                      int src_x,
                                      int src_y,
@@ -417,8 +396,8 @@ cogl_texture_set_region_from_bitmap (CoglTexture *texture,
                                      unsigned int dst_height,
                                      CoglBitmap *bitmap)
 {
-  CoglError *ignore_error = NULL;
-  CoglBool status =
+  GError *ignore_error = NULL;
+  gboolean status =
     _cogl_texture_set_region_from_bitmap (texture,
                                           src_x, src_y,
                                           dst_width, dst_height,
@@ -427,12 +406,11 @@ cogl_texture_set_region_from_bitmap (CoglTexture *texture,
                                           0, /* level */
                                           &ignore_error);
 
-  if (!status)
-    cogl_error_free (ignore_error);
+  g_clear_error (&ignore_error);
   return status;
 }
 
-CoglBool
+gboolean
 _cogl_texture_set_region (CoglTexture *texture,
                           int width,
                           int height,
@@ -442,17 +420,18 @@ _cogl_texture_set_region (CoglTexture *texture,
                           int dst_x,
                           int dst_y,
                           int level,
-                          CoglError **error)
+                          GError **error)
 {
   CoglContext *ctx = texture->context;
   CoglBitmap *source_bmp;
-  CoglBool ret;
+  gboolean ret;
 
-  _COGL_RETURN_VAL_IF_FAIL (format != COGL_PIXEL_FORMAT_ANY, FALSE);
+  g_return_val_if_fail (format != COGL_PIXEL_FORMAT_ANY, FALSE);
+  g_return_val_if_fail (cogl_pixel_format_get_n_planes (format) == 1, FALSE);
 
   /* Rowstride from width if none specified */
   if (rowstride == 0)
-    rowstride = _cogl_pixel_format_get_bytes_per_pixel (format) * width;
+    rowstride = cogl_pixel_format_get_bytes_per_pixel (format, 0) * width;
 
   /* Init source bitmap */
   source_bmp = cogl_bitmap_new_for_data (ctx,
@@ -474,7 +453,7 @@ _cogl_texture_set_region (CoglTexture *texture,
   return ret;
 }
 
-CoglBool
+gboolean
 cogl_texture_set_region (CoglTexture *texture,
 			 int src_x,
 			 int src_y,
@@ -488,12 +467,16 @@ cogl_texture_set_region (CoglTexture *texture,
 			 unsigned int rowstride,
 			 const uint8_t *data)
 {
-  CoglError *ignore_error = NULL;
+  GError *ignore_error = NULL;
   const uint8_t *first_pixel;
-  int bytes_per_pixel = _cogl_pixel_format_get_bytes_per_pixel (format);
-  CoglBool status;
+  int bytes_per_pixel;
+  gboolean status;
+
+  g_return_val_if_fail (format != COGL_PIXEL_FORMAT_ANY, FALSE);
+  g_return_val_if_fail (cogl_pixel_format_get_n_planes (format) == 1, FALSE);
 
   /* Rowstride from width if none specified */
+  bytes_per_pixel = cogl_pixel_format_get_bytes_per_pixel (format, 0);
   if (rowstride == 0)
     rowstride = bytes_per_pixel * width;
 
@@ -509,18 +492,17 @@ cogl_texture_set_region (CoglTexture *texture,
                                      dst_y,
                                      0,
                                      &ignore_error);
-  if (!status)
-    cogl_error_free (ignore_error);
+  g_clear_error (&ignore_error);
   return status;
 }
 
-CoglBool
+gboolean
 cogl_texture_set_data (CoglTexture *texture,
                        CoglPixelFormat format,
                        int rowstride,
                        const uint8_t *data,
                        int level,
-                       CoglError **error)
+                       GError **error)
 {
   int level_width;
   int level_height;
@@ -542,7 +524,7 @@ cogl_texture_set_data (CoglTexture *texture,
                                    error);
 }
 
-static CoglBool
+static gboolean
 get_texture_bits_via_offscreen (CoglTexture *meta_texture,
                                 CoglTexture *sub_texture,
                                 int x,
@@ -557,13 +539,10 @@ get_texture_bits_via_offscreen (CoglTexture *meta_texture,
   CoglOffscreen *offscreen;
   CoglFramebuffer *framebuffer;
   CoglBitmap *bitmap;
-  CoglBool ret;
-  CoglError *ignore_error = NULL;
+  gboolean ret;
+  GError *ignore_error = NULL;
   CoglPixelFormat real_format;
 
-  if (!cogl_has_feature (ctx, COGL_FEATURE_ID_OFFSCREEN))
-    return FALSE;
-
   offscreen = _cogl_offscreen_new_with_texture_full
                                       (sub_texture,
                                        COGL_OFFSCREEN_DISABLE_DEPTH_AND_STENCIL,
@@ -572,7 +551,7 @@ get_texture_bits_via_offscreen (CoglTexture *meta_texture,
   framebuffer = COGL_FRAMEBUFFER (offscreen);
   if (!cogl_framebuffer_allocate (framebuffer, &ignore_error))
     {
-      cogl_error_free (ignore_error);
+      g_error_free (ignore_error);
       return FALSE;
     }
 
@@ -601,8 +580,7 @@ get_texture_bits_via_offscreen (CoglTexture *meta_texture,
                                                    bitmap,
                                                    &ignore_error);
 
-  if (!ret)
-    cogl_error_free (ignore_error);
+  g_clear_error (&ignore_error);
 
   cogl_object_unref (bitmap);
 
@@ -611,7 +589,7 @@ get_texture_bits_via_offscreen (CoglTexture *meta_texture,
   return ret;
 }
 
-static CoglBool
+static gboolean
 get_texture_bits_via_copy (CoglTexture *texture,
                            int x,
                            int y,
@@ -623,17 +601,20 @@ get_texture_bits_via_copy (CoglTexture *texture,
 {
   unsigned int full_rowstride;
   uint8_t *full_bits;
-  CoglBool ret = TRUE;
+  gboolean ret = TRUE;
   int bpp;
   int full_tex_width, full_tex_height;
 
+  g_return_val_if_fail (dst_format != COGL_PIXEL_FORMAT_ANY, FALSE);
+  g_return_val_if_fail (cogl_pixel_format_get_n_planes (dst_format) == 1, FALSE);
+
   full_tex_width = cogl_texture_get_width (texture);
   full_tex_height = cogl_texture_get_height (texture);
 
-  bpp = _cogl_pixel_format_get_bytes_per_pixel (dst_format);
+  bpp = cogl_pixel_format_get_bytes_per_pixel (dst_format, 0);
 
   full_rowstride = bpp * full_tex_width;
-  full_bits = malloc (full_rowstride * full_tex_height);
+  full_bits = g_malloc (full_rowstride * full_tex_height);
 
   if (texture->vtable->get_data (texture,
                                  dst_format,
@@ -654,7 +635,7 @@ get_texture_bits_via_copy (CoglTexture *texture,
   else
     ret = FALSE;
 
-  free (full_bits);
+  g_free (full_bits);
 
   return ret;
 }
@@ -666,8 +647,8 @@ typedef struct
   int orig_height;
   CoglBitmap *target_bmp;
   uint8_t *target_bits;
-  CoglBool success;
-  CoglError *error;
+  gboolean success;
+  GError *error;
 } CoglTextureGetData;
 
 static void
@@ -679,7 +660,8 @@ texture_get_cb (CoglTexture *subtexture,
   CoglTextureGetData *tg_data = user_data;
   CoglTexture *meta_texture = tg_data->meta_texture;
   CoglPixelFormat closest_format = cogl_bitmap_get_format (tg_data->target_bmp);
-  int bpp = _cogl_pixel_format_get_bytes_per_pixel (closest_format);
+  /* We already asserted that we have a single plane format */
+  int bpp = cogl_pixel_format_get_bytes_per_pixel (closest_format, 0);
   unsigned int rowstride = cogl_bitmap_get_rowstride (tg_data->target_bmp);
   int subtexture_width = cogl_texture_get_width (subtexture);
   int subtexture_height = cogl_texture_get_height (subtexture);
@@ -755,7 +737,7 @@ cogl_texture_get_data (CoglTexture *texture,
   int tex_width;
   int tex_height;
   CoglPixelFormat texture_format;
-  CoglError *ignore_error = NULL;
+  GError *ignore_error = NULL;
 
   CoglTextureGetData tg_data;
 
@@ -765,11 +747,14 @@ cogl_texture_get_data (CoglTexture *texture,
   if (format == COGL_PIXEL_FORMAT_ANY)
     format = texture_format;
 
+  /* We only support single plane formats */
+  g_return_val_if_fail (cogl_pixel_format_get_n_planes (format) == 1, 0);
+
   tex_width = cogl_texture_get_width (texture);
   tex_height = cogl_texture_get_height (texture);
 
   /* Rowstride from texture width if none specified */
-  bpp = _cogl_pixel_format_get_bytes_per_pixel (format);
+  bpp = cogl_pixel_format_get_bytes_per_pixel (format, 0);
   if (rowstride == 0)
     rowstride = tex_width * bpp;
 
@@ -780,7 +765,6 @@ cogl_texture_get_data (CoglTexture *texture,
 
   closest_format =
     ctx->texture_driver->find_best_gl_get_data_format (ctx,
-                                                       texture_format,
                                                        format,
                                                        &closest_gl_format,
                                                        &closest_gl_type);
@@ -836,7 +820,7 @@ cogl_texture_get_data (CoglTexture *texture,
                                                         &ignore_error);
       if (!target_bmp)
         {
-          cogl_error_free (ignore_error);
+          g_error_free (ignore_error);
           return 0;
         }
     }
@@ -873,7 +857,7 @@ cogl_texture_get_data (CoglTexture *texture,
     }
   else
     {
-      cogl_error_free (ignore_error);
+      g_error_free (ignore_error);
       tg_data.success = FALSE;
     }
 
@@ -890,8 +874,8 @@ cogl_texture_get_data (CoglTexture *texture,
   if (closest_format != format)
     {
       CoglBitmap *new_bmp;
-      CoglBool result;
-      CoglError *error = NULL;
+      gboolean result;
+      GError *error = NULL;
 
       /* Convert to requested format directly into the user's buffer */
       new_bmp = cogl_bitmap_new_for_data (ctx,
@@ -903,7 +887,7 @@ cogl_texture_get_data (CoglTexture *texture,
 
       if (!result)
         {
-          cogl_error_free (error);
+          g_error_free (error);
           /* Return failure after cleaning up */
           byte_size = 0;
         }
@@ -1091,20 +1075,20 @@ _cogl_texture_set_allocated (CoglTexture *texture,
   _cogl_texture_free_loader (texture);
 }
 
-CoglBool
+gboolean
 cogl_texture_allocate (CoglTexture *texture,
-                       CoglError **error)
+                       GError **error)
 {
   if (texture->allocated)
     return TRUE;
 
   if (texture->components == COGL_TEXTURE_COMPONENTS_RG &&
       !cogl_has_feature (texture->context, COGL_FEATURE_ID_TEXTURE_RG))
-    _cogl_set_error (error,
-                     COGL_TEXTURE_ERROR,
-                     COGL_TEXTURE_ERROR_FORMAT,
-                     "A red-green texture was requested but the driver "
-                     "does not support them");
+    g_set_error (error,
+                 COGL_TEXTURE_ERROR,
+                 COGL_TEXTURE_ERROR_FORMAT,
+                 "A red-green texture was requested but the driver "
+                 "does not support them");
 
   texture->allocated = texture->vtable->allocate (texture, error);
 
@@ -1208,7 +1192,7 @@ void
 cogl_texture_set_components (CoglTexture *texture,
                              CoglTextureComponents components)
 {
-  _COGL_RETURN_IF_FAIL (!texture->allocated);
+  g_return_if_fail (!texture->allocated);
 
   if (texture->components == components)
     return;
@@ -1224,9 +1208,9 @@ cogl_texture_get_components (CoglTexture *texture)
 
 void
 cogl_texture_set_premultiplied (CoglTexture *texture,
-                                CoglBool premultiplied)
+                                gboolean premultiplied)
 {
-  _COGL_RETURN_IF_FAIL (!texture->allocated);
+  g_return_if_fail (!texture->allocated);
 
   premultiplied = !!premultiplied;
 
@@ -1236,7 +1220,7 @@ cogl_texture_set_premultiplied (CoglTexture *texture,
   texture->premultiplied = premultiplied;
 }
 
-CoglBool
+gboolean
 cogl_texture_get_premultiplied (CoglTexture *texture)
 {
   return texture->premultiplied;
